"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var chai_1 = require("chai");
var mu = require("../../../src");
var builders_1 = require("../builders");
var reports_1 = require("../../../test-kit/test-drivers/reports");
var either = mu.either;
describe('List', function () {
    describe('mu instance', function () {
        describe('pop', function () {
            it('should remove the last element from a List', function () {
                var numberList = builders_1.aNumberList();
                var oldArr = numberList.concat();
                var valueRemoved = numberList.pop();
                chai_1.expect(numberList.length).to.equal(oldArr.length - 1);
                chai_1.expect(valueRemoved).to.equal(oldArr.at(oldArr.length - 1));
            });
            it('should return undefined if called on an empty List', function () {
                var numberList = builders_1.anEmptyList();
                var valueRemoved = numberList.pop();
                chai_1.expect(valueRemoved).to.be.undefined;
            });
        });
        describe('push', function () {
            it('it should add a number to a List ', function () {
                var numberList = builders_1.aNumberList([1, 2, 3, 4]);
                var lengthBeforePush = numberList.length;
                var newIndex = numberList.push(5);
                chai_1.expect(newIndex).to.equal(lengthBeforePush + 1);
                chai_1.expect(numberList.at(lengthBeforePush)).to.equal(5);
            });
            it('should add a typed item for non-primitive data (like custom types)', function () {
                var arr = mu.List.of(builders_1.UserType).create([]);
                arr.push({});
                chai_1.expect(arr.at(0)).to.be.instanceOf(builders_1.UserType);
            });
            it('should add a typed item form multiple types if there is _type field', function () {
                var arr = mu.List.of(either(builders_1.UserType, builders_1.AddressType)).create([]);
                arr.push({ _type: 'User' });
                arr.push({ _type: 'Address' });
                chai_1.expect(arr.at(0)).to.be.instanceOf(builders_1.UserType);
                chai_1.expect(arr.at(1)).to.be.instanceOf(builders_1.AddressType);
            });
            it('should support push of multiple items', function () {
                var numberList = builders_1.aNumberList([1, 2]);
                numberList.push(3, 4);
                chai_1.expect(numberList.length).to.equal(4);
                chai_1.expect(numberList.at(2)).to.equal(3);
                chai_1.expect(numberList.at(3)).to.equal(4);
            });
            it("report correct path for field type mismatch in deep field", function () {
                var aList = builders_1.aVeryCompositeContainerList([{}, {}]);
                chai_1.expect(function () { return aList.push({}, { child1: { user: { age: "666" } } }); })
                    .to.report(reports_1.ERROR_FIELD_MISMATCH_IN_LIST_METHOD('push', 'List<VeryCompositeContainer>[3].child1.user.age', 'number', 'string'));
            });
            describe('lifecycleManager', function () {
                var arr, manager, child;
                beforeEach(function () {
                    manager = new LifeCycleManager();
                    arr = mu.List.of(builders_1.UserType).create([new builders.UserType(), new builders.UserType(), new builders.UserType()]);
                    arr.$setManager(manager);
                    child = new builders.UserType();
                    sinon.spy(child, '$setManager');
                });
                if (context.dirtyableElements) {
                    it('sets lifecycle manager in newly added elements', function () {
                        arr.push(child);
                        chai_1.expect(child.$setManager).to.have.been.calledWithExactly(manager);
                    });
                    it('does not try to set lifecycle manager in read-only newly added elements', function () {
                        arr.push(child.$asReadOnly());
                        chai_1.expect(child.$setManager).to.have.not.been.calledWithExactly(manager);
                    });
                }
            });
        });
        describe('set', function () {
            it('should replace an existing element', function () {
                var arr = builders_1.aStringList(['a']);
                arr.set(0, 'b');
                chai_1.expect(arr.toJSON()).to.eql(['b']);
            });
            it('should add an element if none exists', function () {
                var arr = builders_1.anEmptyList();
                arr.set(0, 42);
                chai_1.expect(arr.toJSON()).to.eql([42]);
            });
            it('should return the element', function () {
                var arr = builders_1.aStringList(['a']);
                chai_1.expect(arr.set(0, 'b')).to.eql('b');
            });
            it("report correct path for field type mismatch in deep field", function () {
                var aList = builders_1.aVeryCompositeContainerList([{}, {}]);
                chai_1.expect(function () { return aList.set(1, { child1: { user: { age: "666" } } }); })
                    .to.report(reports_1.ERROR_FIELD_MISMATCH_IN_LIST_METHOD('set', 'List<VeryCompositeContainer>.child1.user.age', 'number', 'string'));
            });
            describe('lifecycleManager', function () {
                var arr, manager, child;
                beforeEach(function () {
                    manager = new LifeCycleManager();
                    arr = mu.List.of(builders_1.UserType).create([new builders.UserType(), new builders.UserType(), new builders.UserType()]);
                    arr.$setManager(manager);
                    child = new builders.UserType();
                    sinon.spy(child, '$setManager');
                });
                if (context.dirtyableElements) {
                    it('sets lifecycle manager in newly added elements', function () {
                        arr.set(0, child);
                        chai_1.expect(child.$setManager).to.have.been.calledWithExactly(manager);
                    });
                    it('does not try to set lifecycle manager in read-only newly added elements', function () {
                        arr.set(0, child.$asReadOnly());
                        chai_1.expect(child.$setManager).to.have.not.been.calledWithExactly(manager);
                    });
                }
            });
        });
        describe('shift', function () {
            it('should return the first element from the List', function () {
                var numberList = builders_1.aNumberList();
                var ListBeforeShift = numberList.concat();
                var valueRemoved = numberList.shift();
                chai_1.expect(ListBeforeShift.at(0)).to.equal(valueRemoved);
            });
            it('should remove an element from the List', function () {
                var numberList = builders_1.aNumberList();
                var lengthBeforeShift = numberList.length;
                numberList.shift();
                chai_1.expect(numberList.length).to.equal(lengthBeforeShift - 1);
            });
        });
        describe('unshift', function () {
            it('should return the length of the List', function () {
                var numberList = builders_1.aNumberList();
                var newLength = numberList.unshift();
                chai_1.expect(numberList.length).to.equal(newLength, 'Did not return the proper List.length');
            });
            it('should add an element to the List', function () {
                var numberList = builders_1.aNumberList();
                var lengthBeforeUnshift = numberList.length;
                numberList.unshift(5);
                chai_1.expect(numberList.length).to.equal(lengthBeforeUnshift + 1);
                chai_1.expect(numberList.at(0)).to.equal(5);
            });
            it("report correct path for field type mismatch in deep field", function () {
                var aList = builders_1.aVeryCompositeContainerList([{}, {}]);
                chai_1.expect(function () { return aList.unshift({}, { child1: { user: { age: "666" } } }); })
                    .to.report(reports_1.ERROR_FIELD_MISMATCH_IN_LIST_METHOD('unshift', 'List<VeryCompositeContainer>[1].child1.user.age', 'number', 'string'));
            });
            describe('lifecycleManager', function () {
                var arr, manager, child;
                beforeEach(function () {
                    manager = new LifeCycleManager();
                    arr = mu.List.of(builders_1.UserType).create([new builders.UserType(), new builders.UserType(), new builders.UserType()]);
                    arr.$setManager(manager);
                    child = new builders.UserType();
                    sinon.spy(child, '$setManager');
                });
                if (context.dirtyableElements) {
                    it('sets lifecycle manager in newly added elements', function () {
                        arr.unshift(child);
                        chai_1.expect(child.$setManager).to.have.been.calledWithExactly(manager);
                    });
                    it('does not try to set lifecycle manager in read-only newly added elements', function () {
                        arr.unshift(child.$asReadOnly());
                        chai_1.expect(child.$setManager).to.have.not.been.calledWithExactly(manager);
                    });
                }
            });
        });
    });
});
//# sourceMappingURL=item-mutations.spec.js.map